classdef LogLik_Factor_Copula
    
    % Date: April - 2020
    % This class contains several log-likelihood functions of 1, 2 and
    % multifactor copula models from the paper
    
    %  "Closed-Form Multi-Factor Copula Models with Observation-Driven Dynamic
    %   Factor Loadings" by Opschoor et al. (2020).
    
    % NOTE: all references to equations refer to Opschoor et al.(2020)
    
    % General input
    %   N: number of assets
    %   T: number of time periods
    %   u_mat: a matrix of (T by N) with PITs
    %   asset_group_vec: a (N by 1) vector classifying each asset to a specific
    %                   industry (e.g. 1 = financials, 2 = health, etc)
    %   ind_t_dist:     1 if t-copula, Gaussian copula otherwise
    %   ind_Rt: optional argument, 1 if you would like to compute R_t, 0 else
    
    % General output
    %   LLF: the MINUS copula log likelihood
    %   loglike_vec: a (T x 1) vector of log-likelihood values at time t
    %   R_vec/ R_mat: fitted dependence matrices (T x N x N)
    %   s_vec/ s_mat: a T by p matrix of scores (with p the number of
    %               unique factor loadings)
    %   f_vec/ f_mat: a T by p matrix of f_values (the unique unscaled factor loadings that
    %               change over time by a score update (with p the number of
    %               unique factor loadings)
    
    
    methods (Static = true)
        
                        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        
        function [LLF,loglike_vec,R_vec,s_vec,f_vec] = LogLik_Copula_1f_eq(N,T,params,u_mat,ind_t_dist);
            
            % 1-factor Copula: 1 equi loading with Normal or t_distribution,
            % see line 12-28 of this function for the general input and output.
            
            % Get out the parameters for the model from the parameter vector params
            % Order: omega_l, A, B [, nu]  (nu only there if ind_t_dist == 1)
            omega_l = params(1);
            A    = params(2);
            B    = params(3);
            
            % Compute the inv cdf (Gaussian or Student's t) of the PITs
            % as input for the copula density later on
            if ind_t_dist ==1
                % Student's t case: important, the parameterization is such that
                % the variance exists, hence the multiplication by sqrt((nu-2)/nu).
                nu      = params(end);
                x_mat_0 = tinv(u_mat,nu);
                x_mat   = sqrt((nu-2)/nu) * x_mat_0;
            else
                % Gaussian case
                x_mat   = norminv(u_mat);
            end
            
            % Initialize the storage objects for the output
            loglike_vec = zeros(T,1);
            f_vec       = zeros(T,1);
            s_vec       = zeros(T,1);
            R_vec       = zeros(T,1);
            
            % initialize fixed matrices required during computations
            i_vec = ones(N,1);
                        
            % Compute the starting value of f(t)
            % For the one-factor-equi model, we target the average
            % sample correlation and then use the parameterization
            % of the loading as rho = f / sqrt(1+f*f)..
            RC_eq       = 1/(N*(N-1))*(i_vec'*corr(x_mat)*i_vec-N);
            f_vec(1)    = sqrt(RC_eq/(1-RC_eq));
                        
            % Loop over all observations to compute the time varying
            % parameter f(t) and implied time-varying correlation
            % parameter.
            for j = 1:T
                
                % Step 0: compute the time varying coefficient and the
                %         data vector at this time
                f_t     = f_vec(j);
                f_2_t   = f_t^2;
                x_j     = x_mat(j,:)';
                
                % Step 1a: compute the factor loading, the idiosyncratic
                %          variance, and some auxiliary quantities
                denom_t                 = 1+f_2_t;	% denominator value, used repeatedly
                lambda_til_prime_vec_t  = f_t/sqrt(denom_t)*i_vec;
                sigma_2_vec_t           = i_vec * 1./ denom_t;
                
                % Step 1b: compute the correlation matrix and its
                %          inverse and determinant
                [R_inv_t,det_R_t] = Admin.Compute_R_t_inv_and_det_R_t(sigma_2_vec_t,lambda_til_prime_vec_t);
                R_t               = Admin.Compute_R_t(sigma_2_vec_t,lambda_til_prime_vec_t);
                % Step 1c: store the time-varying correlation for this model
                R_vec(j,1)        = R_t(1,2);
                
                % CHECKUP: if something wrong with the correlation matrix (not PDS),
                % then return likelihood failure by huge -loglik value.
                if det_R_t <=0 || isnan(det_R_t)
                    loglike_vec = 1e14;
                    break
                else
                    
                    % Step 2: Compute the likelihood for time t=j and store it,
                    %         compute the score function and update the time-varying
                    %         parameter
                    
                    % Step 2a: compute R^{-1} x(t) and x(t)' R^{-1} x(t)
                    coef_R_1    = -0.5;
                    R_inv_x     = R_inv_t * x_j;
                    x_R_inv_x   = x_j' * R_inv_x;
                    
                    % Step 2b: compute the derivative of the copula density with respect to
                    %          R(t) as in (A8)-(A9)
                    if ind_t_dist==1
                        coef_R_2 = 0.5*(nu + N)/(nu-2+ x_R_inv_x );
                    else
                        coef_R_2 = 0.5;
                    end
                    R_2 = R_inv_x * R_inv_x';    % outer product: R^{-1} x(t) x(t)' R^{-1}
                    % dau log c/ dau vec R_t
                    A_mat    = coef_R_1 * R_inv_t + coef_R_2 *R_2;
                    
                    % Step 2c: compute the score, using vec(A_mat)' * vec(B_mat) =
                    %          trace(A_mat' * Bmat) = trace(A_mat * Bmat) given
                    %          the symmetry of A_mat. See (F.3) in the online appendix.
                    g_f = 2 * f_t / (denom_t ^ 2);
                    s_vec(j,1) = g_f * (sum(sum(A_mat,2)) - trace(A_mat));
                    
                    % Step 2d: update the time-varying parameter using the score for the next
                    %          iteration
                    if j<T
                        f_vec(j+1) = omega_l +  A * s_vec(j,1) + B * f_vec(j,1);
                    end
                    
                    % Step 2e: compute the core of the log likelihood for this observation,
                    %          with additional constants added outside the loop, and the
                    %          "marginal" part of the copula density also added outside the
                    %          loop.
                    if ind_t_dist ==1
                        % Student's t copula core
                        loglike_vec(j,1) = -0.5 *(nu+N)*log(1 + x_R_inv_x/(nu-2)) - 0.5*log(det_R_t);
                    else
                        % Gaussian copula core
                        loglike_vec(j,1) = -0.5 * x_R_inv_x - 0.5*log(det_R_t);
                    end
                    
                end
                
            end
            
            % CHECKUP: if the likelihood computations were successful, proceed
            % and return the likelihood value; otherwise, return a failure by
            % returning a high -likelihood value.
            if isreal(loglike_vec) && sum(loglike_vec) ~=1e14
                % complete the likelihood
                if ind_t_dist==1
                    % Student's t case: add the marginal Student t density parts
                    % and the integrating constants
                    loglike_vec = loglike_vec + gammaln(0.5*(nu+N)) + (N-1)*gammaln(nu/2) - N * gammaln(0.5*(nu+1)) ...
                        + 0.5*(nu+1)*sum(log(ones(T,N)+(1/(nu-2))*x_mat.^2),2);
                else
                    % Gaussian case: add the marginal Gaussian density parts
                    loglike_vec = loglike_vec + 0.5*sum(x_mat.^2,2);
                end
                
                % store the -log-likelihood value
                LLF = -sum(loglike_vec);
            else
                LLF = 1e14;
            end
            
        end
        
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        
        function [LLF,loglike_vec,R_mat,s_mat,f_mat] = LogLik_Copula_1f_gr(N,T,params,u_mat,asset_group_vec,ind_t_dist,ind_Rt);
            
            % 1-factor Copula: 1 group loading (hence G different intercepts) with Normal or t_distribution,
            % see line 12-28 of this function for the general input and output.
            
            % Extra input:
            % ind_Rt: 1 if you would like to compute and store Rt, 0 otherwise
            
            % Get out the parameters for the model from the parameter vector params
            % Order: omega_vec, A, B [, nu]  (nu only there if ind_t_dist == 1)
            G           = max(asset_group_vec);
            omega_vec   = params(1:G)';
            A           = params(G+1);
            B           = params(G+2);
            
            % Compute the inv cdf (Gaussian or Student's t) of the PITs
            % as input for the copula density later on
            if ind_t_dist ==1
                % Student's t case: important, the parameterization is such that
                % the variance exists, hence the multiplication by sqrt((nu-2)/nu).
                nu      = params(end);
                x_mat_0 = tinv(u_mat,nu);
                x_mat   = sqrt((nu-2)/nu) * x_mat_0;
            else
                x_mat   = norminv(u_mat);
            end
            
            % Initialize the storage objects for the output
            loglike_vec = zeros(T,1);
            f_mat       = zeros(T,G);
            s_mat       = zeros(T,G);
            if ind_Rt==1
                R_mat = zeros(N,N,T);   %only if ind_Rt = 1,
            else
                R_mat = [];
            end
			
            % initialize fixed matrices required during computations
            i_vec      = ones(N,1);
            S_l_mat    = Admin.Compute_S_l_matrix(asset_group_vec);
            % S_l_mat = (n x G) matrix with entry i,j equal to 1 if asset i belongs to group j.
                        
            % Compute the starting value of f(t)
            % For the one-factor-gr model, we target the average
            % sample correlation and then use the parameterization
            % of the loading as rho = f / sqrt(1+f*f).
            RC_eq       = 1/(N*(N-1))*(i_vec'*corr(x_mat)*i_vec-N);
            f_mat(1,:)  = sqrt(RC_eq/(1-RC_eq))*ones(1,G);
                                    
            
            % Loop over all observations to compute the time varying
            % parameters f(t) and implied time-varying correlation
            % parameters.
            
            for j = 1:T
                
                % Step 0: compute the time varying coefficient and the
                %         data vector at this time
                f_t     = f_mat(j,:);
                f_2_t   = f_t.^2;
                x_j     = x_mat(j,:)';
                
                % Step 1a: compute the factor loading, the idiosyncratic
                %          variance, and some auxiliary quantities
                denom_t                 = 1+f_2_t;	% denominator value, used repeatedly
                lambda_til_prime_t      = f_t./sqrt(denom_t);
                lambda_til_prime_vec_t  = S_l_mat*lambda_til_prime_t';
                sigma_2_vec_t           = S_l_mat* (1./denom_t');
                
                % Step 1b: compute the correlation matrix (if ind_Rt=1) and its
                %          inverse and determinant
                [R_inv_t,det_R_t] = Admin.Compute_R_t_inv_and_det_R_t(sigma_2_vec_t,lambda_til_prime_vec_t);
                if ind_Rt==1
                    R_t         = Admin.Compute_R_t(sigma_2_vec_t,lambda_til_prime_vec_t);
                    R_mat(:,:,j)= R_t;  %store in matrix
                end
                
                % CHECKUP: if something wrong with the correlation matrix (not PDS),
                % then return likelihood failure by huge -loglik value.                
                if det_R_t <=0 || isnan(det_R_t)
                    loglike_vec = 1e14;
                    break
                else
                    
                    % Step 2: Compute the likelihood for time t=j and store it,
                    %         compute the score function and update the time-varying
                    %         parameters
                    
                    % Compute the score is factor-model-specific.
                    
                    % Step 2a: compute R^{-1} x(t) and x(t)' R^{-1} x(t)                    
                    coef_R_1    = -0.5;
                    R_inv_x     = (R_inv_t * x_j);
                    x_R_inv_x   = x_j' * R_inv_x;
                    
                    % Step 2b: compute the derivative of the copula density with respect to
                    %          R(t) as in (A8)-(A9)
                    if ind_t_dist==1
                        coef_R_2 = 0.5*(nu + N)/(nu-2+ x_R_inv_x );
                    else
                        coef_R_2 = 0.5;
                    end                    
                    R_2 = R_inv_x * R_inv_x';  % outer product: R^{-1} x(t) x(t)' R^{-1}
                    
                    % dau_log_c_dau_vec_R
                    A_mat = coef_R_1 * R_inv_t + coef_R_2 *R_2;
                    
                    % Step 2c: compute the score, using  (F.2),(F.3) (F.5) and (F.6) of the online appendix.                    
                    g_f                 =  S_l_mat * diag(1./denom_t.^(3/2));
                    dau_L_tilde_dau_f   =  2* (lambda_til_prime_vec_t' * A_mat) * g_f;
                    dau_D_dau_f         =  S_l_mat * diag((-2*f_t./denom_t.^2));
                    
                    s_mat(j,:) = dau_L_tilde_dau_f +  diag(A_mat)'*dau_D_dau_f;
                    
                    % Step 2d: update the time-varying parameters using the score for the next
                    %          iteration
                    if j<T
                        f_mat(j+1,:) = omega_vec +  A* s_mat(j,:) + B*f_mat(j,:);
                    end
                    
                    % Step 2e: compute the core of the log likelihood for this observation,
                    %          with additional constants added outside the loop, and the
                    %          "marginal" part of the copula density also added outside the
                    %          loop.                                        
                    if ind_t_dist ==1
                        % Student's t copula core
                        loglike_vec(j,1) = -0.5 *(nu+N)*log(1 + x_R_inv_x/(nu-2)) - 0.5*log(det_R_t);
                    else
                        % Gaussian copula core
                        loglike_vec(j,1) = -0.5 * x_R_inv_x - 0.5*log(det_R_t);
                    end
                    
                end
            end
            
            % CHECKUP: if the likelihood computations were successful, proceed
            % and return the likelihood value; otherwise, return a failure by
            % returning a high -likelihood value.            
            if isreal(loglike_vec) && sum(loglike_vec) ~=1e14
                % complete likelihood
                if ind_t_dist==1
                    % Student's t case: add the marginal Student t density parts
                    % and the integrating constants
                    loglike_vec = loglike_vec + gammaln(0.5*(nu+N)) + (N-1)*gammaln(nu/2) - N * gammaln(0.5*(nu+1)) ...
                        + 0.5*(nu+1)*sum(log(ones(T,N)+(1/(nu-2))*x_mat.^2),2);
                else
                    % Gaussian case: add the marginal Gaussian density parts
                    loglike_vec = loglike_vec + 0.5*sum(x_mat.^2,2);
                end
                % store the -log-likelihood value
                LLF = -sum(loglike_vec);
            else
                LLF = 1e14;
            end
            
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        
        function [LLF,loglike_vec,R_mat,s_mat,f_mat] = LogLik_Copula_2f(N,T,params,u_mat,asset_group_vec,ind_t_dist,ind_Rt);
            % Factor Copula, 2 factors, 1 equi factor and 1 factor with group loadings (different intercepts
            % Distribution: Normal or t-copula
            % see line 12-28 of this function for the general input and output.
            
            % EXTRA Input
            % ind_Rt: 1 if you would like to compute and store Rt, 0 otherwise
            
            % specify number of factors (k), number of groups (G) and the number
            % of unique factor loadings (p)
            k = 2;
            G = max(asset_group_vec);
            p = G+1;
            
            % Get out the parameters for the model from the parameter vector params
            % Order: omega_vec, A_vec, B_vec [, nu]  (nu only there if ind_t_dist == 1)
            
            omega_vec   =  params(1:p)';
            A_vec       = [params(p+1) params(p+2)*ones(1,p-1)];
            B_vec       =  params(p+3)*ones(1,p);
            
            % Compute the inv cdf (Gaussian or Student's t) of the PITs
            % as input for the copula density later on
            if ind_t_dist ==1
                nu      = params(end);
                x_mat_0 = tinv(u_mat,nu);
                x_mat   = sqrt((nu-2)/nu) * x_mat_0;
            else
                x_mat   = norminv(u_mat);
            end
            
            % Initialize the storage objects for the output
            loglike_vec = zeros(T,1);
            f_mat       = zeros(T,p);
            s_mat       = zeros(T,p);
            dau_L_tilde_dau_f_mat    = zeros(k*N,p);
            
            if ind_Rt==1
                R_mat = zeros(N,N,T);   %only if ind_Rt = 1,
            else
                R_mat = [];
            end
            
            % initialize fixed matrices required during computations
            i_vec      = ones(N,1);
            S_l_mat    = Admin.Compute_S_l_matrix(asset_group_vec);
            % S_l_mat = (N x G) matrix with entry i,j equal to 1 if asset i belongs to group j.
                        
            % Compute the starting value of f(t)
            % For the 2-factor model, we target the average
            % sample correlation and then use the parameterization
            % of the loading as rho = f / sqrt(1+f*f)..
            RC_eq       = 1/(N*(N-1))*(i_vec'*corr(x_mat)*i_vec-N);
            f_mat(1,:)  = sqrt(RC_eq/(1-RC_eq))*ones(1,p);
                        
            % Loop over all observations to compute the time varying
            % parameters f(t) and implied time-varying correlation
            % parameters.
            
            for j = 1:T
                
                % Step 0: compute the time varying coefficient and the
                %         data vector at this time
                
                f_t_vec   = f_mat(j,:);
                f_t_vec_2 = f_t_vec.^2;
                x_j = x_mat(j,:)';
                
                % Step 1a: compute the factor loading, the idiosyncratic
                %          variance, and some auxiliary quantities
                denom_f_un  = (1+ f_t_vec_2(1) + f_t_vec_2(2:end)');	     % denominator value (Gx1) of UNIQUE f_elements, used repeatedly
                denom_f_vec = (1+ f_t_vec_2(1) + S_l_mat*f_t_vec_2(2:end)'); % denominator value (Nx1), used repeatedly
                denom_f_mat = denom_f_vec*ones(1,k);
                lambda_til_prime_mat_t  = [i_vec*f_t_vec(1) S_l_mat * f_t_vec(2:end)']./ sqrt(denom_f_mat);
                sigma_2_vec_t           = (1./denom_f_vec);
                
                % Step 1b: compute the correlation matrix (if ind_Rt=1) and its
                %          inverse and determinant
                [R_inv_t,det_R_t,Flag]  = Admin.Compute_R_t_inv_and_det_R_t_L_matrix(sigma_2_vec_t,lambda_til_prime_mat_t);
                if ind_Rt==1
                    R_t         = Admin.Compute_R_t(sigma_2_vec_t,lambda_til_prime_mat_t);
                    R_mat(:,:,j)= R_t;
                end
                                
                % CHECKUP: if something wrong with the correlation matrix (not PDS),
                % then return likelihood failure by huge -loglik value.                
                if Flag==1 || isnan(det_R_t)
                    loglike_vec = 1e14;
                    break
                else
                    %Step 2: Compute the likelihood for time t=j and store it,
                    %         compute the score function and update the time-varying
                    %         parameters
                    
                    % Compute the score is factor-model-specific.
                    
                    % Step 2a: compute R^{-1} x(t) and x(t)' R^{-1} x(t)                    
                    coef_R_1    = -0.5;
                    R_inv_x     = (R_inv_t * x_j);
                    x_R_inv_x   =  x_j' * R_inv_x;
                    
                    % Step 2b: compute the derivative of the copula density with respect to
                    %          R(t) as in (A8)-(A9)                    
                    if ind_t_dist==1
                        coef_R_2 = 0.5*(nu + N)/(nu-2+ x_R_inv_x );
                    else
                        coef_R_2 = 0.5;
                    end
                    R_2 = R_inv_x * R_inv_x';  % outer product: R^{-1} x(t) x(t)' R^{-1}
                    
                    % dau_log_c_dau_vec_R
                    A_mat = coef_R_1 * R_inv_t + coef_R_2 *R_2;
                    
                    % Step 2c: compute the score, using  (F.2),(F.3) (F.9) -(F.15) of the online appendix.
                    
                    % auxiliary terms, used repeatedly
                    denom_f_vec_3_2             = denom_f_vec.^(3/2);
                    denom_f_un_3_2              = denom_f_un.^(3/2);
                    
                    % (F.9)
                    dau_lambda_equi_dau_f_equi  =  1./sqrt(denom_f_vec) - i_vec*f_t_vec_2(1)./denom_f_vec_3_2;
                    % (F.10)
                    dau_lambda_gr_dau_f_gr =  S_l_mat * ( diag(1./sqrt(denom_f_un)) - diag(f_t_vec_2(2:end)'./denom_f_un_3_2));
                    % (F.11)
                    dau_lambda_equi_dau_f_gr = S_l_mat * diag( -(f_t_vec(1) *f_t_vec(2:end)')./denom_f_un_3_2);
                    % (F.12)
                    dau_lambda_gr_dau_f_equi =  S_l_mat * ((-f_t_vec(1) * f_t_vec(2:end)')./denom_f_un_3_2);
                    
                    % (F.13) and (F.14)
                    dau_L_tilde_dau_f_mat(1:k:end,:) = [dau_lambda_equi_dau_f_equi dau_lambda_equi_dau_f_gr];
                    dau_L_tilde_dau_f_mat(2:k:end,:) = [dau_lambda_gr_dau_f_equi dau_lambda_gr_dau_f_gr];
                    
                    % (F.15)
                    dau_D_dau_f_mat = -2*[f_t_vec(1)./denom_f_vec.^2 S_l_mat*diag(f_t_vec(2:end)'./denom_f_un.^2)];
                    
                    % combining above with (F.2) and (F.3)
                    L_A         = lambda_til_prime_mat_t' * A_mat;
                    s_vec_a     = 2* (L_A(:))'* dau_L_tilde_dau_f_mat;
                    s_vec_b     = diag(A_mat)' * dau_D_dau_f_mat;
                    s_mat(j,:)  = s_vec_a + s_vec_b;
                    
                    % Step 2d: update the time-varying parameters using the score for the next
                    %          iteration
                    if j<T
                        f_mat(j+1,:) = omega_vec +  A_vec.*s_mat(j,:) + B_vec.*f_mat(j,:);
                    end
                    % Step 2e: compute the core of the log likelihood for this observation,
                    %          with additional constants added outside the loop, and the
                    %          "marginal" part of the copula density also added outside the
                    %          loop.                                        
                    if ind_t_dist ==1
                        % Student's t copula core
                        loglike_vec(j,1) = -0.5 *(nu+N)*log(1 + x_R_inv_x/(nu-2)) - 0.5*log(det_R_t);
                    else
                        % Gaussian copula core
                        loglike_vec(j,1) = -0.5 * x_R_inv_x - 0.5*log(det_R_t);
                    end
                end
                
                
            end
            
            % CHECKUP: if the likelihood computations were successful, proceed
            % and return the likelihood value; otherwise, return a failure by
            % returning a high -likelihood value.                        
            if isreal(loglike_vec) && sum(loglike_vec) ~=1e14
                % complete likelihood
                if ind_t_dist==1
                    % Student's t case: add the marginal Student t density parts
                    % and the integrating constants                    
                    loglike_vec = loglike_vec + gammaln(0.5*(nu+N)) + (N-1)*gammaln(nu/2) - N * gammaln(0.5*(nu+1)) ...
                        + 0.5*(nu+1)*sum(log(ones(T,N)+(1/(nu-2))*x_mat.^2),2);
                else
                    % Gaussian case: add the marginal Gaussian density parts
                    loglike_vec = loglike_vec + 0.5*sum(x_mat.^2,2);
                end
                % store the -log-likelihood value
                LLF = -sum(loglike_vec);
            else
                LLF = 1e14;
            end
            
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        
        function [LLF,loglike_vec,R_mat,s_mat,f_mat] = LogLik_Copula_MF(N,T,params,u_mat,asset_group_vec,ind_t_dist,ind_Rt);
            % MF Copula, 1 equi factor and G industry specific factors
            % Distribution: Normal or t-copula
            % see line 12-28 of this function for the general input and output.
            
            % EXTRA Input
            % ind_Rt: 1 if you would like to compute and store Rt, 0 otherwise
            
            % specify number of factors (k), number of groups (G) and the number
            % of unique factor loadings (p)            
            G       = max(asset_group_vec);
            k       = G + 1;
            p       = k;
            
            % Get out the parameters for the model from the parameter vector params
            % Order: omega_vec, A_vec, B [, nu]  (nu only there if ind_t_dist == 1)            
            omega_vec   = params(1:p)';
            A_vec       = [params(p+1) params(p+2)*ones(1,p-1)];
            B           = params(p+3);
            
            % Compute the inv cdf (Gaussian or Student's t) of the PITs
            % as input for the copula density later on
            if ind_t_dist ==1
                nu      = params(end);
                x_mat_0 = tinv(u_mat,nu);
                x_mat   = sqrt((nu-2)/nu) * x_mat_0;
            else
                x_mat   = norminv(u_mat);
            end
            
            % Initialize the storage objects for the output
            loglike_vec = zeros(T,1);
            f_mat       = zeros(T,p);
            s_mat       = zeros(T,p);
            dau_L_tilde_dau_f_mat    = zeros(k*N,p);            
            if ind_Rt ==1
                R_mat = zeros(N,N,T);
            else
                R_mat = [];
            end
            
            % initialize fixed matrices required during computations
            i_vec      = ones(N,1);
            S_l_mat    = Admin.Compute_S_l_matrix(asset_group_vec);
            % S_l_mat = (n x G) matrix with entry i,j equal to 1 if asset i belongs to group j.
                        
            % Compute the starting value of f(t)
            % For the MF model, we target the average
            % sample correlation and then use the parameterization
            % of the loading as rho = f / sqrt(1+f*f)..
            RC_eq       = 1/(N*(N-1))*(i_vec'*corr(x_mat)*i_vec-N);
            f_mat(1,:)  = sqrt(RC_eq/(1-RC_eq))*ones(1,p);
                        
            % Loop over all observations to compute the time varying
            % parameters f(t) and implied time-varying correlation
            % parameters.
            
            for j = 1:T
                
                % Step 0: compute the time varying coefficient and the
                %         data vector at this time
                f_t_vec     = f_mat(j,:);
                f_t_vec_2   = f_t_vec.^2;
                x_j         = x_mat(j,:)';
                
                % Step 1a: compute the factor loading, the idiosyncratic
                %          variance, and some auxiliary quantities
                denom_f_un  = (1+ f_t_vec_2(1) + f_t_vec_2(2:end)');	     % denominator value (Gx1) of UNIQUE f_elements, used repeatedly
                denom_f_vec = (1+ f_t_vec_2(1) + S_l_mat*f_t_vec_2(2:end)'); % denominator value (Nx1), used repeatedly
                denom_f_mat     = denom_f_vec*ones(1,k);
                lambda_til_prime_mat_t = [i_vec*f_t_vec(1) S_l_mat * diag(f_t_vec(2:end))]./ sqrt(denom_f_mat);
                sigma_2_vec_t   = (1./denom_f_vec);
                
                % Step 1b: compute the correlation matrix (if ind_Rt=1) and its
                %          inverse and determinant
                [R_inv_t,det_R_t,Flag] = Admin.Compute_R_t_inv_and_det_R_t_L_matrix(sigma_2_vec_t,lambda_til_prime_mat_t);
                if ind_Rt ==1
                    R_t         = Admin.Compute_R_t(sigma_2_vec_t,lambda_til_prime_mat_t);
                    R_mat(:,:,j)= R_t;
                end
                
                % CHECKUP: if something wrong with the correlation matrix (not PDS),
                % then return likelihood failure by huge -loglik value.                
                if Flag==1 || isnan(det_R_t)
                    loglike_vec = 1e14;
                    break
                else
                    %Step 2: Compute the likelihood for time t=j and store it,
                    %         compute the score function and update the time-varying
                    %         parameters
                    
                    % Compute the score is factor-model-specific.
                    
                    % Step 2a: compute R^{-1} x(t) and x(t)' R^{-1} x(t)                    
                    coef_R_1    = -0.5;
                    R_inv_x     = (R_inv_t * x_j);
                    x_R_inv_x   = x_j' * R_inv_x;
                    
                    % Step 2b: compute the derivative of the copula density with respect to
                    %          R(t) as in (A8)-(A9)                    
                    if ind_t_dist==1
                        coef_R_2 = 0.5*(nu + N)/(nu-2+ x_R_inv_x );
                    else
                        coef_R_2 = 0.5;
                    end
                    R_2 = R_inv_x * R_inv_x';  % outer product: R^{-1} x(t) x(t)' R^{-1}
                    
                    % dau_log_c_dau_vec_R
                    A_mat = coef_R_1 * R_inv_t + coef_R_2 *R_2;
                    
                    % Step 2c: compute the score, using  (F.2),(F.3) (F.9)-(F.11) (F.15)-(F.18)
                    % of the online appendix.
                    
                    % auxiliary terms used repeatedly
                    denom_f_vec_3_2         = denom_f_vec.^(3/2);
                    denom_f_un_3_2          = denom_f_un.^(3/2);
                    
                    %(F.9)
                    dau_lambda_equi_dau_f_equi  =  1./sqrt(denom_f_vec) - i_vec*f_t_vec_2(1)./denom_f_vec_3_2;
                    %(F.10)
                    dau_lambda_gr_dau_f_gr 	=  S_l_mat * ( diag(1./sqrt(denom_f_un)) - diag(f_t_vec_2(2:end)'./denom_f_un_3_2));
                    %(F.11)
                    dau_lambda_equi_dau_f_gr 	= S_l_mat * diag( -(f_t_vec(1) *f_t_vec(2:end)')./denom_f_un_3_2);
                    
                    %(F.16)
                    dau_L_tilde_dau_f_mat(1:k:end,:) = [dau_lambda_equi_dau_f_equi dau_lambda_equi_dau_f_gr];
                    
                    %(F.17) and (F.18)
                    for i = 1:k-1
                        dau_L_tilde_dau_f_mat(i+1:k:end,[1 i+1]) = [dau_lambda_equi_dau_f_gr(:,i) dau_lambda_gr_dau_f_gr(:,i)];
                    end
                    
                    %  (F.15)
                    dau_D_dau_f_mat = -2*[f_t_vec(1)*i_vec S_l_mat*diag(f_t_vec(2:end))]./(denom_f_vec.^2 * ones(1,p));
                                                            
                    % combining above with (F.2) and (F.3)
                    L_A        = lambda_til_prime_mat_t' * A_mat;
                    s_vec_a    = 2* (L_A(:))'* dau_L_tilde_dau_f_mat;
                    s_vec_b    = diag(A_mat)' *dau_D_dau_f_mat;
                    s_mat(j,:) = s_vec_a + s_vec_b;
                    
                    % Step 2d: update the time-varying parameters using the score for the next
                    %          iteration                    
                    if j<T
                        f_mat(j+1,:) = omega_vec +  A_vec.*s_mat(j,:) + B*f_mat(j,:);
                    end
                    % Step 2e: compute the core of the log likelihood for this observation,
                    %          with additional constants added outside the loop, and the
                    %          "marginal" part of the copula density also added outside the
                    %          loop.
                    
                    if ind_t_dist ==1
                        % Student's t copula core
                        loglike_vec(j,1) = -0.5 *(nu+N)*log(1 + x_R_inv_x/(nu-2)) - 0.5*log(det_R_t);
                    else
                        % Gaussian copula core
                        loglike_vec(j,1) = -0.5 * x_R_inv_x - 0.5*log(det_R_t);
                    end
                end
                
                
            end
            % CHECKUP: if the likelihood computations were successful, proceed
            % and return the likelihood value; otherwise, return a failure by
            % returning a high -likelihood value.
            
            
            if isreal(loglike_vec) && sum(loglike_vec) ~=1e14
                % complete log likelihood
                if ind_t_dist==1
                    % Student's t case: add the marginal Student t density parts
                    % and the integrating constants
                    loglike_vec = loglike_vec + gammaln(0.5*(nu+N)) + (N-1)*gammaln(nu/2) - N * gammaln(0.5*(nu+1)) ...
                        + 0.5*(nu+1)*sum(log(ones(T,N)+(1/(nu-2))*x_mat.^2),2);
                else
                    % Gaussian case: add the marginal Gaussian density parts
                    loglike_vec = loglike_vec + 0.5*sum(x_mat.^2,2);
                end
                % store the -log-likelihood value
                LLF = -sum(loglike_vec);
            else
                
                LLF = 1e14;
            end
            
        end
                        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        function [dist_sq,L_tilde_prime_L_tilde] = Moment_based_omega_LT_FC(G,params, rho_vec_sample_bl_vech);
            
            % This function contains the first step of the two-step
            % approach when estimating the MF-LT model with a scaled loading
            % matrix defined in eq(6). This is a moment based estimator of f_bar.
            
            % The function computes L_m of eq (12), using eq(11).
            
            % Input
            % rho_vec_sample_bl_vech:  a (vech) of the G times G matrix R_M defined below eq(11),
            %                          containing the average within/between industry correlation.
            % G: number of groups (industries)
            % params: parameters
            
            % Output
            % L_m: the squared distance
            % L_tilde_prime_L_tilde: matrix of off-diagonal correlations,
            % see eq(12)
            
            g_vec_cum            = cumsum(G:-1:1);
            omega_vec            = params;
            
            f_t_vec       = omega_vec';
            f_t_vec_2     = f_t_vec.^2;
            
            % build the model implied correlation matrix Rm of eq (12)
            teller_i = 1;
            denom_f_vec_g = ones(G,1);
            f_prime_mat_t = zeros(G,G);
            for i = 1:G
                g_i_c              = g_vec_cum(i);
                denom_f_vec_g      = denom_f_vec_g + [zeros(i-1,1); f_t_vec_2(teller_i:g_i_c)'];
                f_prime_mat_t(:,i) = [zeros(i-1,1); f_t_vec(teller_i:g_i_c)'];
                teller_i = g_i_c+1;
            end
            
            denom_f_mat_g            = denom_f_vec_g*ones(1,G);
            L_tilde_prime_mat_t      = f_prime_mat_t./sqrt(denom_f_mat_g);
            L_tilde_prime_L_tilde    = L_tilde_prime_mat_t* L_tilde_prime_mat_t';
            vech_FC                  = Admin.Sym2Vech(G,L_tilde_prime_L_tilde);

			% compute the Frobenius norm of the difference between the sample
			% moments and the model-implied moments
            dist_vec = vech_FC - rho_vec_sample_bl_vech;
            dist_sq  = dist_vec'*dist_vec;
            
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
                        
        function [LLF,loglike_vec,R_mat,s_mat,f_mat] = LogLik_Copula_LT_factor_given_omega(N,T,params,f_hat_vec,u_mat,asset_group_vec,n_vec,ind_t_dist,ind_Rt);
            
            % MF_LT copula, (lower triangular matrix) g factors, see eq (6)
            % see line 12-28 of this function for the general input and output.
            
            % Extra input:
            % ind_Rt: 1 if you would like to compute and store the block correlation matrix, 0 otherwise
            % f_hat_vec: estimated f_bar using the first step moment
            %            estimator using eq(12).
            % n_vec    : G by 1 vector where entry g (g = 1,..,G) denotes
            % the number of assets in group g.
                        
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            % IMPORTANT NOTE:
            % asset_group_vec and u_mat needs to be SORTED before running this function (!!!)
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            
            % specify number of factors (k), number of groups (G) and the number
            % of unique factor loadings (p)            
            G = max(asset_group_vec);
            g_vec_cum            = cumsum(G:-1:1);
            k                    = G;
            p                    = G*(G+1)/2;
            
            % Get out the parameters for the model from the parameter vector params
            % Order: A_vec, B_vec [, nu]  (nu only there if ind_t_dist == 1)           
            A_vec  = params(1);
            B_vec  = params(2);
            omega_vec = (1-B_vec)*f_hat_vec';
            
            % Compute the inv cdf (Gaussian or Student's t) of the PITs
            % as input for the copula density later on
            if ind_t_dist ==1
                nu      = params(end);
                x_mat_0 = tinv(u_mat,nu);
                x_mat   = sqrt((nu-2)/nu) * x_mat_0;
            else
                x_mat   = norminv(u_mat);
            end
            
            % Initialize the storage objects for the output
            loglike_vec         = zeros(T,1);            
            f_mat               = zeros(T,p);
            s_mat               = zeros(T,p);
            if ind_Rt==1
                R_mat   = zeros(G,G,T);
            else
                R_mat   = [];
            end
            
            % Compute the starting value of f(t)
            % For the MF_LT model, we set f(1,:)
            % f_bar, estimated by moment-based estimator, see eq(11) and eq(12)            
            f_mat(1,:) = f_hat_vec;
            
            % Maka a new asset_group_vec that is ascending from 1 to G,
            % where 1 is your first sorted industry, 2 the second etc. 
            
            teller_0 = 1;
            asset_group_vec_new = zeros(N,1);
            for m = 1:G
                asset_group_vec_new(teller_0:teller_0+n_vec(m)-1) = m;
                teller_0                                          = teller_0 + n_vec(m);
            end
            
            % initialize fixed matrices required during computations
            S_l_mat    = Admin.Compute_S_l_matrix(asset_group_vec_new);
                        
            % Part of the score dau_c/dau_vec_f_t is build outside the time-loop
            % this is the selection matrix "dau_L_tilde_dau_lambda_tilde".
            
            % Consider eq(6); we have G(G+1)/2 different f_values.
            % Since lambda_tilde_1,1,t and lambda_tilde_2,2,t are in the
            % same row, they will both appear in the non-linear scaling
            % formulas of (3).
            
            % Here we assign which lambda_tilde values belong to the same
            % row of L_tilde_prime (or COLUMN of L_tilde)
            Ind_lambda_to_group_vec = [];
            for i = 1:G
                Ind_lambda_to_group_vec = [Ind_lambda_to_group_vec; [i:G]'];
            end                                   
            Ind_lambda_to_group_mat      = [[1:p]' Ind_lambda_to_group_vec];
            
            % we are now ready to build "dau_L_tilde_dau_lambda_tilde"
            dau_L_tilde_dau_lambda_tilde = zeros(N*k,p);
            
             
            ind_match_lambda_group_cell = cell(G,1);
            
            %     Building dau_L_tilde_dau_lambda_tilde consists of the following steps:            
            %     1) Loop over each column i of L_tilde (i = 1:G)
            %     (note that lambda_tilde_i is the first element in each colomn)
            %     2) compute dau_L_t_lambda_tilde_i;
            %     3) check which lambda_tilde_j (j = 1,...,p) are in the same column i of L_tilde
            %        store these into a vector ind_lambda_group_i (and in a
            %        cell ind_match_lambda_group_cell)
            %     4) if i > 1 we have more than on lambda_tilde_j in the
            %        i-th column of L_tilde. More speficically, we have exactly (i-1) other
            %        lambda_tilde_j parameters.
            %        4a) loop over these (i-1) other lambda_tilde_j values (using the vector ind_lambda_g_group_i) 
            %            and compute dau_L_t_lambda_tilde_j
            %     5) return to step 1).
                        
            for i =  1:G
                % step 2
                dau_L_tilde_dau_lambda_tilde(sum(n_vec(1:i-1))*G+1:G:sum(n_vec(1:i))*G,i) =1;
                
                % step 3
                ind_lambda_g_group_i           = Ind_lambda_to_group_mat(Ind_lambda_to_group_mat(:,2)==i,1);
                ind_match_lambda_group_cell{i} = ind_lambda_g_group_i;
                
                % step 4
                for j = 1:i-1
                    lambda_j                   = ind_lambda_g_group_i(j+1);
                    dau_L_tilde_dau_lambda_tilde(sum(n_vec(1:i-1))*G+1+j:G:sum(n_vec(1:i))*G+j,lambda_j) =1;
                end
            end
            
            
            % Loop over all observations to compute the time varying
            % parameters f(t) and implied time-varying correlation
            % parameters.
            
            
            for j = 1:T
                
                % Step 0: compute the time varying coefficient and the
                %         data vector at this time
                
                f_t_vec   = f_mat(j,:);
                f_t_vec_2 = f_t_vec.^2;
                x_j       = x_mat(j,:)';
                                
                % Build the matrix L_til_prime_mat_t. See (F.19) of the web
                % appendix.                 
                % We also build a second matrix L_tilde_prime_mat_tg (F.19)
                % with N_g = 1 for all values of g)                 
                % This G x G matrix is based on only unique elements of f_j,t.
                % We (could) use this later on to build the G x G block correlation
                % matrix with within and between group correlations.
                
                teller_i        = 1;
                denom_f_un      = ones(G,1);
                denom_f_vec     = ones(N,1);
                f_prime_mat_t   = zeros(N,G);
                f_prime_mat_tg  = zeros(G,G);
                for i = 1:k
                    g_i_c           = g_vec_cum(i);
                    denom_f_vec     = denom_f_vec + S_l_mat(:,i:G)*f_t_vec_2(teller_i:g_i_c)';  % Based on (F.19) and used repeatedly. 
                    denom_f_un      = denom_f_un + [zeros(i-1,1); f_t_vec_2(teller_i:g_i_c)'];  % This is (F.21) and used repeatedly.
                    f_prime_mat_t(:,i)  = S_l_mat(:,i:G)*f_t_vec(teller_i:g_i_c)';
                    f_prime_mat_tg(:,i) = [zeros(i-1,1); f_t_vec(teller_i:g_i_c)'];
                    teller_i            = g_i_c+1;
                end
                
                % Step 1b: compute the correlation matrix (if ind_Rt=1) and its
                %          inverse and determinant
                denom_f_mat             = denom_f_vec*ones(1,k);
                lambda_til_prime_mat_t  = f_prime_mat_t./ sqrt(denom_f_mat);
                sigma_2_vec_t           = (1./denom_f_vec);
                [R_inv_t,det_R_t,Flag] = Admin.Compute_R_t_inv_and_det_R_t_L_matrix(sigma_2_vec_t,lambda_til_prime_mat_t);
                if ind_Rt==1
                    denom_f_mat_tg          = denom_f_un*ones(1,G);
                    L_tilde_prime_mat_tg    = f_prime_mat_tg./sqrt(denom_f_mat_tg);
                    Rt_Block                = L_tilde_prime_mat_tg* L_tilde_prime_mat_tg';
                    R_mat(:,:,j) = Rt_Block;
                end
                                                
                % CHECKUP: if something wrong with the correlation matrix (not PDS),
                % then return likelihood failure by huge -loglik value.                
                
                if Flag==1 || isnan(det_R_t)
                    loglike_vec = 1e14;
                    break
                else
                    %Step 2: Compute the likelihood for time t=j and store it,
                    %         compute the score function and update the time-varying
                    %         parameters
                    
                    % Compute the score is factor-model-specific.
                    
                    % Step 2a: compute R^{-1} x(t) and x(t)' R^{-1} x(t)                    
                    coef_R_1    = -0.5;
                    R_inv_x     = (R_inv_t * x_j);
                    x_R_inv_x   =  x_j' * R_inv_x;
                    
                    % Step 2b: compute the derivative of the copula density with respect to
                    %          R(t) as in (A8)-(A9)                    
                    if ind_t_dist==1
                        coef_R_2 = 0.5*(nu + N)/(nu-2+ x_R_inv_x );
                    else
                        coef_R_2 = 0.5;
                    end
                    
                    R_2 = R_inv_x * R_inv_x';  % outer product: R^{-1} x(t) x(t)' R^{-1}
                    
                    % dau_log_c_dau_vec_R
                    A_mat = coef_R_1 * R_inv_t + coef_R_2 *R_2;
                    
                    % Step 2c: compute the score, using  (F.2),(F.3) (F.20)-(F.25)
                    % of the online appendix.
                    
                    % auxilary term used repeatedly 
                    denom_f_un_3_2           = denom_f_un.^(3/2);
                    
                    % (F.22) for j = 1,2,..., G(G+1/2)
                    dau_lambda_tilde_dau_f_vec  = 1./sqrt(denom_f_un(Ind_lambda_to_group_mat(:,2))) - f_t_vec_2'./denom_f_un_3_2(Ind_lambda_to_group_mat(:,2));
                    dau_lambda_tilde_dau_f_mat  = diag(dau_lambda_tilde_dau_f_vec);
                    
                    dau_sigma2_dau_f_mat         = zeros(G,p);
                    
                    % fill dau_lambda_tilde_dau_f and dau_sigma2_dau_f
                                        
                    % The loop below computes (F.23) for all values of j, looping over all G columns. 
                    % That is: check which values of f_(t,j) are in column g. Then compute the cross deratives.                     
                     
                    % The loop also computes (F.25) for each g (g = 1,...,G) 
                    % That is: check which values of f_(t,j) are in column g. 
                    % Then compute (F.25) w.r.t. to all these f_(t,j) values.                      
                    for g = 1:G
                        % find out which f_(t,j) are in column g
                        ind_cross_lambda_g =  ind_match_lambda_group_cell{g};
                        if g>1
                            for m = 1:g
                                ind_2   = ind_cross_lambda_g;
                                ind_m1  = ind_cross_lambda_g(m);
                                ind_2(m)= [];
                                
                                %(F.23)
                                dau_lambda_tilde_i_dau_m_vec             = -(f_t_vec(ind_m1).*f_t_vec(ind_2))./denom_f_un_3_2(g);
                                dau_lambda_tilde_dau_f_mat(ind_m1,ind_2) = dau_lambda_tilde_i_dau_m_vec;
                            end
                        end
                        %(F.25)
                        dau_sigma2_dau_f_mat(g,ind_cross_lambda_g) = -2*f_t_vec(ind_cross_lambda_g)./ (denom_f_un(g)^2);
                    end
                    
                    %(F.20)
                    dau_L_tilde_dau_f_mat =  dau_L_tilde_dau_lambda_tilde * dau_lambda_tilde_dau_f_mat;
                    
                    
                    % (F.24)                                         
                    dau_D_dau_f_mat = S_l_mat *  dau_sigma2_dau_f_mat;
                    
                    % combining above with (F.2) and (F.3)
                    L_A         = lambda_til_prime_mat_t' * A_mat;
                    s_vec_a     = 2* (L_A(:))'* dau_L_tilde_dau_f_mat;
                    s_vec_b     = diag(A_mat)' * dau_D_dau_f_mat;
                    s_mat(j,:)  = s_vec_a + s_vec_b;
                    
                    % Step 2d: update the time-varying parameters using the score for the next
                    %          iteration
                    if j<T
                        f_mat(j+1,:) = omega_vec +  A_vec.*s_mat(j,:) + B_vec.*f_mat(j,:);
                    end
                    
                    % Step 2e: compute the core of the log likelihood for this observation,
                    %          with additional constants added outside the loop, and the
                    %          "marginal" part of the copula density also added outside the
                    %          loop.                    
                    if ind_t_dist ==1
                        % Student's t copula core
                        loglike_vec(j,1) = -0.5 *(nu+N)*log(1 + x_R_inv_x/(nu-2)) - 0.5*log(det_R_t);
                    else
                        % Gaussian copula core
                        loglike_vec(j,1) = -0.5 * x_R_inv_x - 0.5*log(det_R_t);
                    end
                end
                
                
            end
            % CHECKUP: if the likelihood computations were successful, proceed
            % and return the likelihood value; otherwise, return a failure by
            % returning a high -likelihood value.
            
            
            if isreal(loglike_vec) && sum(loglike_vec) ~=1e14
                % complete likelihood
                if ind_t_dist==1
                    % Student's t case: add the marginal Student t density parts
                    % and the integrating constants                    
                    loglike_vec = loglike_vec + gammaln(0.5*(nu+N)) + (N-1)*gammaln(nu/2) - N * gammaln(0.5*(nu+1)) ...
                        + 0.5*(nu+1)*sum(log(ones(T,N)+(1/(nu-2))*x_mat.^2),2);
                else
                    % Gaussian case: add the marginal Gaussian density parts
                    loglike_vec = loglike_vec + 0.5*sum(x_mat.^2,2);
                end
                % store the -log-likelihood value
                LLF = -sum(loglike_vec);
            else
                LLF = 1e14;
            end
            
            
        end        
                
    end
end